from nostr.key import PublicKey, PrivateKey
from . import bech32


def from_nip19(nip19string: str):
    """
    Decode nip-19 formatted string into:
    private key, public key, event id or profile public key
    """
    hrp, data, spec = bech32.bech32_decode(nip19string)
    data = bech32.convertbits(data, 5, 8)
    if hrp == 'npub':
        return PublicKey(bytes(data[:-1]))
    elif hrp == 'nsec':
        return PrivateKey(bytes(data[:-1]))
    elif hrp in ('nevent', 'nprofile'):
        tlv = {0: [], 1: []}
        while data:
            t = data[0]
            try:
                l = data[1]
            except IndexError:
                break
            v = data[2:2+l]
            data = data[2+l:]
            if not v:
                continue
            tlv[t].append(v)
        key_or_id = bytes(tlv[0][0]).hex()
        relays = []
        for relay in tlv[1]:
            relays.append(bytes(relay).decode('utf8'))
        return hrp, key_or_id, relays


def to_nip19(ntype: str, payload: str, relays=None):
    """
    Encode object as nip-19 compatible string
    """
    if ntype in ('npub', 'nsec'):
        data = bytes.fromhex(payload)
    elif ntype in ('nprofile', 'nevent'):
        # payload is event id
        event_id = bytes.fromhex(payload)
        data = bytearray()
        data.append(0)
        data.append(len(event_id))
        data.extend(event_id)
        if relays:
            for r in relays:
                r = r.encode()
                data.append(1)
                data.append(len(r))
                data.extend(r)
    converted_bits = bech32.convertbits(data, 8, 5)
    return bech32.bech32_encode(ntype, converted_bits, bech32.Encoding.BECH32)



"""Top-level package for aionostr."""

__author__ = """Dave St.Germain"""
__email__ = 'dave@st.germa.in'
__version__ = '0.6.0'

import time
from .relay import Manager, Relay


async def get_anything(anything:str, relays=None, verbose=False, stream=False, origin='aionostr'):
    """
    Return anything from the nostr network
    anything: event id, nprofile, nevent, npub, nsec, or query

    To stream events, set stream=True. This will return an asyncio.Queue to
    retrieve events from
    """
    from .util import from_nip19

    query = None
    single_event = False
    if isinstance(anything, list):
        if anything[0] == 'REQ':
            query = anything[2]
        else:
            raise NotImplementedError(anything)
    elif isinstance(anything, dict):
        query = anything
    elif anything.strip().startswith('{'):
        from json import loads
        query = loads(anything)
    elif anything.startswith(('nprofile', 'nevent', 'npub', 'nsec', 'nostr:')):
        anything = anything.replace('nostr:', '', 1)
        obj = from_nip19(anything)
        if not isinstance(obj, tuple):
            return obj.hex()
        else:
            relays = obj[2] or relays
            if obj[0] == 'nprofile':
                query = {"kinds": [0], "authors": [obj[1]]}
            else:
                query = {"ids": [obj[1]]}
                single_event = True
    else:
        query = {"ids": [anything]}
        single_event = True
    if verbose:
        import sys
        sys.stderr.write(f"Retrieving {query} from {relays}\n")
    if query:
        if not relays:
            raise NotImplementedError("No relays to use")

        man = Manager(relays, verbose=verbose, origin=origin)
        if not stream:
            async with man:
                return [event async for event in man.get_events(query, single_event=single_event, only_stored=True)]
        else:
            import asyncio
            queue = asyncio.Queue()
            async def _task():
                async with man:
                    async for event in man.get_events(query, single_event=single_event, only_stored=False):
                        await queue.put(event)
            asyncio.create_task(_task())
            return queue


async def add_event(relays, event:dict=None, private_key='', kind=1, pubkey='', content='', created_at=None, tags=None, direct_message='', verbose=False):
    """
    Add an event to the network, using the given relays
    event can be specified (as a dict)
    or will be created from the passed in parameters
    """
    if not event:
        from nostr.key import PrivateKey
        from .event import Event
        from .util import from_nip19
        created_at = created_at or int(time.time())
        tags = tags or []
        if not private_key:
            raise Exception("Missing private key")

        if private_key.startswith('nsec'):
            private_key = from_nip19(private_key).hex()
        prikey = PrivateKey(bytes.fromhex(private_key))

        if not pubkey:
            pubkey = prikey.public_key.hex()
        if direct_message:
            dm_pubkey = from_nip19(direct_message).hex() if direct_message.startswith('npub') else direct_message
            tags.append(['p', dm_pubkey])
            kind = 4
            content = prikey.encrypt_message(content, dm_pubkey)
        event = Event(pubkey=pubkey, content=content, created_at=created_at, tags=tags, kind=kind)
        event.sign(prikey.hex())
        event_id = event.id
    else:
        event_id = event['id']
    async with Manager(relays, verbose=verbose) as man:
        if private_key:
            await man.authenticate(private_key)
        await man.add_event(event)
    return event_id


async def add_events(relays, event_iterator):
    async with Manager(relays) as man:
        for event in event_iterator:
            await man.add_event(event)


"""Console script for aionostr."""
import sys
import click
import asyncio
import time
import os
from functools import wraps
from . import get_anything, add_event

DEFAULT_RELAYS = os.getenv('NOSTR_RELAYS', 'wss://nostr-pub.wellorder.net,wss://brb.io,wss://relay.damus.io').split(',')


def async_cmd(func):
  @wraps(func)
  def wrapper(*args, **kwargs):
    return asyncio.run(func(*args, **kwargs))
  return wrapper


@click.group()
def main(args=None):
    """Console script for aionostr."""
    # click.echo("Replace this message by putting your code into "
    #            "aionostr.cli.main")
    # click.echo("See click documentation at https://click.palletsprojects.com/")
    return 0


@main.command()
@click.option('-r', 'relays', help='relay url', multiple=True, default=DEFAULT_RELAYS)
@click.option('-s', '--stream', help='stream results', is_flag=True, default=False)
@click.option('-v', '--verbose', help='verbose results', is_flag=True, default=False)
@click.option('-q', '--query', help='query json')
@click.option('--ids', help='ids')
@click.option('--authors', help='authors')
@click.option('--kinds', help='kinds')
@click.option('--etags', help='etags')
@click.option('--ptags', help='ptags')
@click.option('--since', help='since')
@click.option('--until', help='until')
@click.option('--limit', help='limit')
@async_cmd
async def query(ids, authors, kinds, etags, ptags, since, until, limit, query, relays, stream, verbose):
    """
    Run a query once and print events
    """
    import json
    if not sys.stdin.isatty():
        query = json.loads(sys.stdin.readline())
    elif query:
        query = json.loads(query)
    else:
        query = {}
        if ids:
            query['ids'] = ids.split(',')
        if authors:
            query['authors'] = authors.split(',')
        if kinds:
            query['kinds'] = [int(k) for k in kinds.split(',')]
        if etags:
            query['#e'] = etags.split(',')
        if ptags:
            query['#p'] = ptags.split(',')
        if since:
            query['since'] = int(since)
        if until:
            query['until'] = int(until)
        if limit:
            query['limit'] = int(limit)
    if not query:
        click.echo("some type of query is required")
        return -1
    await _get(query, relays, verbose=verbose, stream=stream)



async def _get(anything, relays, verbose=False, stream=False):
    response = await get_anything(anything, relays, verbose=verbose, stream=stream)
    if isinstance(response, str):
        click.echo(response)
    elif isinstance(response, asyncio.Queue):
        while True:
            event = await response.get()
            click.echo(event)
    else:
        for event in response:
            click.echo(event)


@main.command()
@click.argument("anything")
@click.option('-r', 'relays', help='relay url', multiple=True, default=DEFAULT_RELAYS)
@click.option('-v', '--verbose', help='verbose results', is_flag=True, default=False)
@async_cmd
async def get(anything, relays, verbose, stream=False):
    """
    Get any nostr event
    """
    await _get(anything, relays, verbose=verbose, stream=stream)


@main.command()
@click.option('-r', 'relays', help='relay url', multiple=True, default=DEFAULT_RELAYS)
@click.option('-v', '--verbose', help='verbose results', is_flag=True, default=False)
@click.option('--content', default='', help='content')
@click.option('--kind', default=20000, help='kind', type=int)
@click.option('--created', default=int(time.time()), type=int, help='created_at')
@click.option('--pubkey', default='', help='public key')
@click.option('--tags', default='[]', help='tags')
@click.option('--private-key', default='', help='private key')
@click.option('--dm', default='', help='pubkey to send dm')
@async_cmd
async def send(content, kind, created, tags, pubkey, relays, private_key, dm, verbose):
    """
    Send an event to the network

    private key can be set using environment variable NOSTR_KEY
    """
    import json
    from .util import to_nip19
    tags = json.loads(tags)
    private_key = private_key or os.getenv('NOSTR_KEY', '')
    if not sys.stdin.isatty():
        event = json.loads(sys.stdin.readline())
    else:
        event = None
    event_id = await add_event(
        relays,
        event=event,
        pubkey=pubkey,
        private_key=private_key,
        created_at=int(created),
        kind=kind,
        content=content,
        tags=tags,
        direct_message=dm,
        verbose=verbose,
    )
    click.echo(event_id)
    click.echo(to_nip19('nevent', event_id, relays))


@main.command()
@click.argument("anything")
@click.option('-r', 'relays', help='relay url', multiple=True, default=DEFAULT_RELAYS)
@click.option('-v', '--verbose', help='verbose results', is_flag=True, default=False)
@click.option('-t', '--target', help='target relay', required=True)
@async_cmd
async def mirror(anything, relays, target, verbose):
    """
    Mirror a query from source relays to the target relay
    """
    if verbose:
        click.echo(f'mirroring: {anything} from {relays} to {target}')
    from . import Manager
    async with Manager([target]) as man:
        result_queue = await get_anything(anything, relays=relays, verbose=verbose, stream=True)
        count = 0
        while True:
            event = await result_queue.get()
            await man.add_event(event, check_response=True)
            count += 1
            if verbose:
                click.echo(f'{event.id} from {event.pubkey}')
            else:
                if count % 100 == 0:
                    click.echo(f'{count}...')
        click.echo(f'{count} sent')


@main.command()
@click.argument("ntype")
@click.argument("obj_id")
@click.option('-r', 'relays', help='relay url', multiple=True, default=DEFAULT_RELAYS)
def make_nip19(ntype, obj_id, relays):
    """
    Create nip-19 string for given object id
    """
    from .util import to_nip19
    obj = to_nip19(ntype, obj_id, relays=relays)
    click.echo(obj)


@main.command()
def gen():
    """
    Generate a private/public key pair
    """
    from nostr.key import PrivateKey
    from .util import to_nip19

    pk = PrivateKey()
    click.echo(to_nip19('nsec', pk.hex()))
    click.echo(to_nip19('npub', pk.public_key.hex()))


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



"""
forked from https://github.com/jeffthibault/python-nostr.git
"""
import time
import functools
from enum import IntEnum
from secp256k1 import PrivateKey, PublicKey
from hashlib import sha256

try:
    import rapidjson
    loads = rapidjson.loads
    dumps = functools.partial(rapidjson.dumps, ensure_ascii=False)
except ImportError:
    import json
    loads = json.loads
    dumps = functools.partial(json.dumps, separators=(',', ':'), ensure_ascii=False)


class EventKind(IntEnum):
    SET_METADATA = 0
    TEXT_NOTE = 1
    RECOMMEND_RELAY = 2
    CONTACTS = 3
    ENCRYPTED_DIRECT_MESSAGE = 4
    DELETE = 5


class Event:
    def __init__(
            self,
            pubkey: str='', 
            content: str='', 
            created_at: int=0, 
            kind: int=EventKind.TEXT_NOTE, 
            tags: "list[list[str]]"=[], 
            id: str=None,
            sig: str=None) -> None:
        if not isinstance(content, str):
            raise TypeError("Argument 'content' must be of type str")
        
        self.pubkey = pubkey
        self.content = content
        self.created_at = created_at or int(time.time())
        self.kind = int(kind)
        self.tags = tags
        self.sig = sig
        if not id:
            id = Event.compute_id(self.pubkey, self.created_at, self.kind, self.tags, self.content)
        self.id = id

    @property
    def id_bytes(self):
        return bytes.fromhex(self.id)

    @property
    def is_ephemeral(self):
        return self.kind >= 20000 and self.kind < 30000

    @property
    def is_replaceable(self):
        return self.kind >= 10000 and self.kind < 20000

    @staticmethod
    def serialize(public_key: str, created_at: int, kind: int, tags: "list[list[str]]", content: str) -> bytes:
        data = [0, public_key, created_at, kind, tags, content]
        data_str = dumps(data)
        return data_str.encode()

    @staticmethod
    def compute_id(public_key: str, created_at: int, kind: int, tags: "list[list[str]]", content: str) -> str:
        return sha256(Event.serialize(public_key, created_at, kind, tags, content)).hexdigest()

    @staticmethod
    def from_tuple(row):
        tags = row[4]
        if isinstance(tags, str):
            tags = loads(tags)
        return Event(
            id=row[0].hex(),
            created_at=row[1],
            kind=row[2],
            pubkey=row[3].hex(),
            tags=tags,
            sig=row[5].hex(),
            content=row[6],
        )

    def sign(self, private_key_hex: str) -> None:
        sk = PrivateKey(bytes.fromhex(private_key_hex))
        sig = sk.schnorr_sign(bytes.fromhex(self.id), None, raw=True)
        self.sig = sig.hex()

    def verify(self) -> bool:
        try:
            pub_key = PublicKey(bytes.fromhex("02" + self.pubkey), True) # add 02 for schnorr (bip340)
        except Exception as e:
            return False
        event_id = Event.compute_id(self.pubkey, self.created_at, self.kind, self.tags, self.content)
        verified = pub_key.schnorr_verify(bytes.fromhex(event_id), bytes.fromhex(self.sig), None, raw=True)
        for tag in self.tags:
            if tag[0] == 'delegation':
                # verify delegation signature
                _, delegator, conditions, sig = tag
                to_sign = (':'.join(['nostr', 'delegation', self.pubkey, conditions])).encode('utf8')
                delegation_verified = PublicKey(
                    bytes.fromhex("02" + delegator),
                    True
                ).schnorr_verify(
                    sha256(to_sign).digest(),
                    bytes.fromhex(sig),
                    None,
                    raw=True
                )
                if not delegation_verified:
                    return False
        return verified

    def to_tuple(self):
        return (
            self.id,
            self.pubkey,
            self.created_at,
            self.kind,
            dumps(self.tags),
            self.content,
            self.sig
        )

    def __str__(self):
        return dumps(self.to_json_object())

    def to_json_object(self) -> dict:
        return {
            "id": self.id,
            "pubkey": self.pubkey,
            "created_at": self.created_at,
            "kind": self.kind,
            "tags": self.tags,
            "content": self.content,
            "sig": self.sig
        }



import asyncio
import secrets
import time
import sys
from contextlib import asynccontextmanager
from collections import defaultdict, namedtuple
from websockets import connect, exceptions
from .event import Event

try:
    from rapidjson import dumps, loads
except ImportError:
    from json import dumps, loads

if hasattr(asyncio, 'timeout'):
    timeout = asyncio.timeout
else:
    # python < 3.11 does not have asyncio.timeout
    # rather than re-implement it, we'll just do nothing
    @asynccontextmanager
    async def timeout(duration):
        yield


Subscription = namedtuple('Subscription', ['filters','queue'])


class Relay:
    """
    Interact with a relay
    """
    def __init__(self, url, verbose=False, origin:str = ''):
        self.url = url
        self.ws = None
        self.receive_task = None
        self.subscriptions = defaultdict(lambda: Subscription(filters=[], queue=asyncio.Queue()))
        self.event_adds = asyncio.Queue()
        self.notices = asyncio.Queue()
        self.challenge = None
        self.verbose = verbose
        self.origin = origin or url

    async def connect(self, retries=5):
        for i in range(retries):
            try:
                self.ws = await connect(self.url, origin=self.origin)
            except:
                await asyncio.sleep(0.2 * i)
            else:
                break
        else:
            raise Exception(f"Cannot connect to {self.url}")
        if self.receive_task is None:
            self.receive_task = asyncio.create_task(self._receive_messages())
        await asyncio.sleep(0.01)

    async def reconnect(self):
        await self.connect(20)
        for sub_id, sub in self.subscriptions.items():
            if self.verbose:
                sys.stderr.write(f'resubscribing to {sub.filters}')
            await self.send(["REQ", sub_id, *sub.filters])

    async def close(self):
        self.receive_task.cancel()
        await self.ws.close()

    async def _receive_messages(self):
        while True:
            try:
                async with timeout(30.0):
                    message = await self.ws.recv()

                if self.verbose:
                    sys.stderr.write(message + '\n')
                message = loads(message)
                if message[0] == 'EVENT':
                    await self.subscriptions[message[1]].queue.put(Event(**message[2]))
                elif message[0] == 'EOSE':
                    await self.subscriptions[message[1]].queue.put(None)
                elif message[0] == 'OK':
                    await self.event_adds.put(message)
                elif message[0] == 'NOTICE':
                    await self.notices.put(message[1])
                elif message[0] == 'AUTH':
                    self.challenge = message[1]
                else:
                    sys.stderr.write(message)
            except asyncio.CancelledError:
                return
            except exceptions.ConnectionClosedError:
                await self.reconnect()
            except asyncio.TimeoutError:
                continue
            except:
                import traceback; traceback.print_exc()

    async def send(self, message):
        try:
            await self.ws.send(dumps(message))
        except exceptions.ConnectionClosedError:
            await self.reconnect()
            await self.ws.send(dumps(message))

    async def add_event(self, event, check_response=False):
        if isinstance(event, Event):
            event = event.to_json_object()
        await self.send(["EVENT", event])
        if check_response:
            response = await self.event_adds.get()
            return response[1]

    async def subscribe(self, sub_id: str, *filters, queue=None):
        self.subscriptions[sub_id] = Subscription(filters=filters, queue=queue or asyncio.Queue())
        await self.send(["REQ", sub_id, *filters])
        return self.subscriptions[sub_id].queue

    async def unsubscribe(self, sub_id):
        await self.send(["CLOSE", sub_id])
        del self.subscriptions[sub_id]

    async def authenticate(self, private_key:str):
        if not self.challenge:
            return False
        from nostr.key import PrivateKey
        pk = PrivateKey(bytes.fromhex(private_key))
        auth_event = Event(
            kind=22242,
            pubkey=pk.public_key.hex(),
            tags=[
                ['challenge', self.challenge],
                ['relay', self.url]
            ]
        )
        auth_event.sign(pk.hex())
        await self.send(["AUTH", auth_event.to_json_object()])
        await asyncio.sleep(0.1)
        return True

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, ex_type, ex, tb):
        await self.close()


class Manager:
    """
    Manage a collection of relays
    """
    def __init__(self, relays=None, verbose=False, origin='aionostr'):
        self.relays = [Relay(r, verbose=verbose, origin=origin) for r in (relays or [])]
        self.subscriptions = {}

    def add(self, url):
        self.relays.append(Relay(url))

    async def monitor_queues(self, queues, output):
        seen = set()
        num = len(queues)
        num_eose = 0
        while True:
            get_funcs = [queue.get() for queue in queues]
            for func in asyncio.as_completed(get_funcs):
                result = await func
                if result:
                    eid = result.id_bytes
                    if eid not in seen:
                        await output.put(result)
                        seen.add(eid)
                else:
                    num_eose += 1
                    if num_eose == num:
                        await output.put(result)

    async def broadcast(self, func, *args, **kwargs):
        results = []
        for relay in self.relays:
            results.append(await getattr(relay, func)(*args, **kwargs))
        return results

    async def connect(self):
        await self.broadcast('connect')

    async def close(self):
        await self.broadcast('close')

    async def add_event(self, event, check_response=False):
        await self.broadcast('add_event', event, check_response=check_response)

    async def authenticate(self, private_key:str):
        await self.broadcast('authenticate', private_key)

    async def subscribe(self, sub_id: str, *filters):
        queues = []
        for relay in self.relays:
            queues.append(await relay.subscribe(sub_id, *filters))
        queue = asyncio.Queue()
        self.subscriptions[sub_id] = asyncio.create_task(self.monitor_queues(queues, queue))
        return queue

    async def unsubscribe(self, sub_id):
        await self.broadcast('unsubscribe', sub_id)
        self.subscriptions[sub_id].cancel()
        del self.subscriptions[sub_id]

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, ex_type, ex, tb):
        await self.close()

    async def get_events(self, *filters, only_stored=True, single_event=False):
        sub_id = secrets.token_hex(4)
        queue = await self.subscribe(sub_id, *filters)
        while True:
            event = await queue.get()
            if event is None:
                if only_stored:
                    break
            else:
                yield event
                if single_event:
                    break
        await self.unsubscribe(sub_id)


